# Default multikit "invoked command" for directories
#  The program watches the network for incoming entries,
#  and also posts new entries (from the local multikit) to the net
#  version: 1998.02.27

# The main routine (invoked command):
proc mk_default_dir_sharer {progId ourInternetAddress ourPortNum ourTTL ourPublicKey args} {
	# Initialize the network:
	global ourSocket ourDirId
	if {[catch {groupsock_create $ourInternetAddress $ourPortNum $ourTTL} socket]} {
		return
	}
	set ourSocket $socket
	set ourDirId $progId

	global estimatedBW lastDataSeenTime
	set estimatedBW 0
	set lastDataSeenTime 0

	# Watch for incoming data:
	fileevent $ourSocket readable handlePacketFromNet

	# Start 'self-describing' this directory:
	reinput $progId SELF
}

# Accept input from "multikit", and share it on the network.
proc multikit_input {particCode incarnation ttl
		     dirId progId parentId expirDate args} {
	global MAFP_version
	global ourSocket

	# The current "on the wire" protocol (MAFP) is very simple:
	# Just send the "dirId", etc., arguments as a Tcl string,
	# with the protocol version, command code, and
	# data incarnation in front:

	if {![info exists ourSocket]} return
	set dataToWrite [concat $MAFP_version $particCode $incarnation \
				$dirId $progId $parentId $expirDate $args]
	groupsock_write $ourSocket $dataToWrite $ttl

	# Do this again after a delay:
	set dataSize [string length $dataToWrite]
	set delayMilliseconds [computeReinputDelay $ttl $dataSize]
	after $delayMilliseconds reinput $dirId $progId
}

# Clean up state, prior to exiting.
proc multikit_cleanup {dirId} {
	global ourSocket
	groupsock_delete $ourSocket
}

proc reinput {dirId progId} {
	# First, make sure our directory is still outputting this item,
	# and, if so, get the most up-to-date version of its state.
	set currentParams [multikit_getSharingParams $dirId $progId]

	# Then, reinvoke the input procedure:
	if {$currentParams != {}} {
		eval multikit_input $currentParams
	}
}

# Version of the "on the wire" protocol
set MAFP_version 3

# Read and process an incoming packet from the net:
proc handlePacketFromNet {} {
	global ourSocket
	set packetData [groupsock_read $ourSocket]

	##### TEMP: We currently understand only the simple, Tcl-compatible on-the-wire protocol:
	global MAFP_version
	set version [lindex $packetData 0]
	if {[string compare $version $MAFP_version] != 0} return

	# Perform sanity checks on the data.  We explicitly check only that
	# the 'command' is "d" ('display'), "p" ('post'), or "x" ('delete')
	# - the only commands we currently implement - and that the directory
	# being added to (or self-described) is ours.
	# Other checks will be performed inside the multikit_input call itself:
	set cmd [lindex $packetData 1]
	if {[string first $cmd "dpx"] < 0} return

	set dirId [lindex $packetData 3]
	global ourDirId
	if {[string compare $dirId $ourDirId] != 0} return

	tallyIncomingData [string length $packetData]

	catch {eval multikit_output [lrange $packetData 1 end]}
}

proc computeReinputDelay {ttl dataSize} {
	# Try not to exceed a bandwidth of 5000/($ttl+1) bytes/sec
	# (This is comparable to the bandwidth limits used by SAP)
	set desiredBW [expr {5000/($ttl+1)}]

	# Figure out how long we'd need to wait before sending the data,
	# in order to get an average bandwidth, over 10 minutes, of $desiredBW
	# However, impose a minimum wait of 10 seconds, and a maximum of 20 minutes.
	global estimatedBW
	if {$estimatedBW == 0} {
		set delay [expr {int( $dataSize/$desiredBW) }]
	} else {
		set delay [expr {int( ($dataSize - 600*($desiredBW-$estimatedBW)) / $estimatedBW )}]
	}
	if {$delay < 10} {
		set delay 10
	} elseif {$delay > 1200} {
		set delay 1200
	}

	# Choose a random time, in milliseconds, between 0.5*$delay and 1.5*$delay:
	set delay [randomBetween [expr {$delay*500}] [expr {$delay*1500}]]
	return $delay
}

proc tallyIncomingData {dataSize} {
	# Use this record of incoming data to reestimate the average bandwidth
	# (in bytes/sec).  Compute a weighted average over 10 minutes (600 secs).
	global estimatedBW lastDataSeenTime
	set timeNow [clock seconds]
	set timeSinceLast [expr {$timeNow - $lastDataSeenTime}]
	if {$timeSinceLast > 600} {set timeSinceLast 600}
	set estimatedBW \
	    [expr {($estimatedBW*(600-$timeSinceLast) + $dataSize) / 600.0}]
	set lastDataSeenTime $timeNow
}
